@@ -1081,6 +1081,7 @@ class AmphoraRepository(BaseRepository):
                 id=amphora_id).first()
             load_balancer.amphorae.append(amphora)
 
+    @oslo_db_api.wrap_db_retry(max_retries=5, retry_on_deadlock=True)
     def allocate_and_associate(self, session, load_balancer_id):
         """Allocate an amphora for a load balancer.
 
@@ -1473,6 +1474,10 @@ class L7RuleRepository(BaseRepository):
         with session.begin(subtransactions=True):
             if not model_kwargs.get('id'):
                 model_kwargs.update(id=uuidutils.generate_uuid())
+            if model_kwargs.get('l7policy_id'):
+                l7policy_db = session.query(models.L7Policy).filter_by(
+                    id=model_kwargs.get('l7policy_id')).first()
+                model_kwargs.update(l7policy=l7policy_db)
             l7rule = self.model_class(**model_kwargs)
             validate.l7rule_data(l7rule)
             session.add(l7rule)
@@ -1611,6 +1616,10 @@ class L7PolicyRepository(BaseRepository):
                 pool_db = session.query(models.Pool).filter_by(
                     id=model_kwargs.get('redirect_pool_id')).first()
                 model_kwargs.update(redirect_pool=pool_db)
+            if model_kwargs.get('listener_id'):
+                listener_db = session.query(models.Listener).filter_by(
+                    id=model_kwargs.get('listener_id')).first()
+                model_kwargs.update(listener=listener_db)
             l7policy = self.model_class(
                 **validate.sanitize_l7policy_api_args(model_kwargs,
                                                       create=True))
