@@ -21,6 +21,8 @@ from oslo_utils import uuidutils
 from octavia.common import constants
 import octavia.common.context
 from octavia.common import data_models
+from octavia.common import exceptions
+from octavia.tests.common import constants as c_const
 from octavia.tests.functional.api.v2 import base
 
 
@@ -46,6 +48,19 @@ class TestPool(base.BaseAPITest):
         self.listener_id = self.listener.get('id')
 
         self.set_lb_status(self.lb_id)
+        self._setup_udp_lb_resources()
+
+    def _setup_udp_lb_resources(self):
+        self.udp_lb = self.create_load_balancer(uuidutils.generate_uuid()).get(
+            'loadbalancer')
+        self.udp_lb_id = self.udp_lb.get('id')
+        self.set_lb_status(self.udp_lb_id)
+
+        self.udp_listener = self.create_listener(
+            constants.PROTOCOL_UDP, 8888,
+            self.udp_lb_id).get('listener')
+        self.udp_listener_id = self.udp_listener.get('id')
+        self.set_lb_status(self.udp_lb_id)
 
     def test_get(self):
         api_pool = self.create_pool(
@@ -124,21 +139,16 @@ class TestPool(base.BaseAPITest):
         self.conf.config(group='api_settings', auth_strategy=auth_strategy)
         self.assertEqual(self.NOT_AUTHORIZED_BODY, response.json)
 
-    def test_get_hides_deleted(self):
+    def test_get_deleted_gives_404(self):
         api_pool = self.create_pool(
             self.lb_id,
             constants.PROTOCOL_HTTP,
             constants.LB_ALGORITHM_ROUND_ROBIN,
             listener_id=self.listener_id).get(self.root_tag)
 
-        response = self.get(self.POOLS_PATH)
-        objects = response.json.get(self.root_tag_list)
-        self.assertEqual(len(objects), 1)
         self.set_object_status(self.pool_repo, api_pool.get('id'),
                                provisioning_status=constants.DELETED)
-        response = self.get(self.POOLS_PATH)
-        objects = response.json.get(self.root_tag_list)
-        self.assertEqual(len(objects), 0)
+        self.get(self.POOL_PATH.format(pool_id=api_pool.get('id')), status=404)
 
     def test_bad_get(self):
         self.get(self.POOL_PATH.format(pool_id=uuidutils.generate_uuid()),
@@ -156,6 +166,22 @@ class TestPool(base.BaseAPITest):
         self.assertEqual(1, len(pools))
         self.assertEqual(api_pool.get('id'), pools[0].get('id'))
 
+    def test_get_all_hides_deleted(self):
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            listener_id=self.listener_id).get(self.root_tag)
+
+        response = self.get(self.POOLS_PATH)
+        objects = response.json.get(self.root_tag_list)
+        self.assertEqual(len(objects), 1)
+        self.set_object_status(self.pool_repo, api_pool.get('id'),
+                               provisioning_status=constants.DELETED)
+        response = self.get(self.POOLS_PATH)
+        objects = response.json.get(self.root_tag_list)
+        self.assertEqual(len(objects), 0)
+
     def test_get_all_admin(self):
         project_id = uuidutils.generate_uuid()
         lb1 = self.create_load_balancer(uuidutils.generate_uuid(), name='lb1',
@@ -472,9 +498,24 @@ class TestPool(base.BaseAPITest):
         pools = self.get(self.POOLS_PATH, params={
             'fields': ['id', 'project_id']}).json
         for pool in pools['pools']:
-            self.assertIn(u'id', pool.keys())
-            self.assertIn(u'project_id', pool.keys())
-            self.assertNotIn(u'description', pool.keys())
+            self.assertIn(u'id', pool)
+            self.assertIn(u'project_id', pool)
+            self.assertNotIn(u'description', pool)
+
+    def test_get_one_fields_filter(self):
+        pool1 = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            name='pool1').get(self.root_tag)
+        self.set_lb_status(lb_id=self.lb_id)
+
+        pool = self.get(
+            self.POOL_PATH.format(pool_id=pool1.get('id')),
+            params={'fields': ['id', 'project_id']}).json.get(self.root_tag)
+        self.assertIn(u'id', pool)
+        self.assertIn(u'project_id', pool)
+        self.assertNotIn(u'description', pool)
 
     def test_get_all_filter(self):
         po1 = self.create_pool(
@@ -483,6 +524,10 @@ class TestPool(base.BaseAPITest):
             constants.LB_ALGORITHM_ROUND_ROBIN,
             name='pool1').get(self.root_tag)
         self.set_lb_status(lb_id=self.lb_id)
+        hm = self.create_health_monitor(po1['id'],
+                                        constants.HEALTH_MONITOR_HTTP,
+                                        1, 1, 1, 1).get('healthmonitor')
+        self.set_lb_status(lb_id=self.lb_id)
         self.create_pool(
             self.lb_id,
             constants.PROTOCOL_HTTP,
@@ -497,7 +542,7 @@ class TestPool(base.BaseAPITest):
         self.set_lb_status(lb_id=self.lb_id)
 
         pools = self.get(self.POOLS_PATH, params={
-            'id': po1['id']}).json
+            'id': po1['id'], 'healthmonitor_id': hm['id']}).json
         self.assertEqual(1, len(pools['pools']))
         self.assertEqual(po1['id'],
                          pools['pools'][0]['id'])
@@ -686,6 +731,120 @@ class TestPool(base.BaseAPITest):
             **optionals).get(self.root_tag)
         self.assertEqual(self.project_id, api_pool.get('project_id'))
 
+    def test_create_udp_case_source_ip(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0'}
+        api_pool = self.create_pool(
+            None,
+            constants.PROTOCOL_UDP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            listener_id=self.udp_listener_id,
+            session_persistence=sp).get(self.root_tag)
+        self.assertEqual(constants.PROTOCOL_UDP, api_pool.get('protocol'))
+        self.assertEqual(constants.LB_ALGORITHM_ROUND_ROBIN,
+                         api_pool.get('lb_algorithm'))
+        self.assertEqual(constants.SESSION_PERSISTENCE_SOURCE_IP,
+                         api_pool.get('session_persistence')['type'])
+        self.assertEqual(3, api_pool.get(
+            'session_persistence')['persistence_timeout'])
+        self.assertEqual('255.255.255.0', api_pool.get(
+            'session_persistence')['persistence_granularity'])
+        self.assertIsNone(api_pool.get(
+            'session_persistence')['cookie_name'])
+        self.assert_correct_status(
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id,
+            pool_id=api_pool.get('id'),
+            lb_prov_status=constants.PENDING_UPDATE,
+            listener_prov_status=constants.PENDING_UPDATE,
+            pool_prov_status=constants.PENDING_CREATE,
+            pool_op_status=constants.OFFLINE)
+
+    def test_negative_create_udp_case(self):
+        # Error create pool with udp protocol but non-udp-type
+        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+              "cookie_name": 'test-cookie-name'}
+        req_dict = {
+            'listener_id': self.udp_listener_id,
+            'protocol': constants.PROTOCOL_UDP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        expect_error_msg = ("Validation failure: Cookie names are not "
+                            "supported for %s pools.") % constants.PROTOCOL_UDP
+        res = self.post(self.POOLS_PATH, self._build_body(req_dict),
+                        status=400, expect_errors=True)
+        self.assertEqual(expect_error_msg, res.json['faultstring'])
+        self.assert_correct_status(
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
+        # Error create pool with any non-udp-types and udp session persistence
+        # options.
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0'}
+        req_dict = {
+            'listener_id': self.udp_listener_id,
+            'protocol': constants.PROTOCOL_UDP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': None}
+        for type in [constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+                     constants.SESSION_PERSISTENCE_APP_COOKIE]:
+            expect_error_msg = ("Validation failure: Session persistence of "
+                                "type %s is not supported for %s protocol "
+                                "pools.") % (type, constants.PROTOCOL_UDP)
+            sp.update({'type': type})
+            req_dict['session_persistence'] = sp
+            res = self.post(self.POOLS_PATH, self._build_body(req_dict),
+                            status=400,
+                            expect_errors=True)
+            self.assertEqual(expect_error_msg, res.json['faultstring'])
+            self.assert_correct_status(
+                lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
+        # Error create pool with source ip session persistence and wrong
+        # options.
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0',
+              "cookie_name": 'test-cookie-name'}
+        req_dict = {
+            'listener_id': self.udp_listener_id,
+            'protocol': constants.PROTOCOL_UDP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        expect_error_msg = (
+            "Validation failure: session_persistence %s type for %s "
+            "protocol only accepts: type, persistence_timeout, "
+            "persistence_granularity.") % (
+            constants.SESSION_PERSISTENCE_SOURCE_IP, constants.PROTOCOL_UDP)
+        res = self.post(self.POOLS_PATH, self._build_body(req_dict),
+                        status=400, expect_errors=True)
+        self.assertEqual(expect_error_msg, res.json['faultstring'])
+        self.assert_correct_status(
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
+        # Error create non-udp pool with udp session persistence
+        sps = [{"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+                "persistence_timeout": 3,
+                "persistence_granularity": '255.255.255.0'},
+               {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
+                "persistence_timeout": 3,
+                "persistence_granularity": '255.255.255.0'}]
+        req_dict = {
+            'listener_id': self.listener_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN}
+        expect_error_msg = ("Validation failure: persistence_timeout and "
+                            "persistence_granularity is only for %s protocol "
+                            "pools.") % constants.PROTOCOL_UDP
+        for s in sps:
+            req_dict.update({'session_persistence': s})
+            res = self.post(self.POOLS_PATH, self._build_body(req_dict),
+                            status=400, expect_errors=True)
+            self.assertEqual(expect_error_msg, res.json['faultstring'])
+            self.assert_correct_status(
+                lb_id=self.lb_id, listener_id=self.listener_id)
+
     def test_bad_create(self):
         pool = {'name': 'test1'}
         self.post(self.POOLS_PATH, self._build_body(pool), status=400)
@@ -714,20 +873,19 @@ class TestPool(base.BaseAPITest):
             'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN}
         self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
 
-    def test_create_with_bad_handler(self):
-        self.handler_mock().pool.create.side_effect = Exception()
-        api_pool = self.create_pool(
-            self.lb_id,
-            constants.PROTOCOL_HTTP,
-            constants.LB_ALGORITHM_ROUND_ROBIN,
-            listener_id=self.listener_id).get(self.root_tag)
-        self.assert_correct_status(
-            lb_id=self.lb_id, listener_id=self.listener_id,
-            pool_id=api_pool.get('id'),
-            lb_prov_status=constants.ACTIVE,
-            listener_prov_status=constants.ACTIVE,
-            pool_prov_status=constants.ERROR,
-            pool_op_status=constants.OFFLINE)
+    @mock.patch('octavia.api.drivers.utils.call_provider')
+    def test_create_with_bad_provider(self, mock_provider):
+        mock_provider.side_effect = exceptions.ProviderDriverError(
+            prov='bad_driver', user_msg='broken')
+        lb_pool = {
+            'loadbalancer_id': self.lb_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'project_id': self.project_id}
+        response = self.post(self.POOLS_PATH, self._build_body(lb_pool),
+                             status=500)
+        self.assertIn('Provider \'bad_driver\' reports error: broken',
+                      response.json.get('faultstring'))
 
     def test_create_over_quota(self):
         self.start_quota_mock(data_models.Pool)
@@ -757,7 +915,7 @@ class TestPool(base.BaseAPITest):
         self.set_lb_status(self.lb_id)
         response = self.get(self.POOL_PATH.format(
             pool_id=api_pool.get('id'))).json.get(self.root_tag)
-        self.assertNotEqual('new_name', response.get('name'))
+        self.assertEqual('new_name', response.get('name'))
         self.assertIsNotNone(response.get('created_at'))
         self.assertIsNotNone(response.get('updated_at'))
         self.assert_correct_status(
@@ -809,7 +967,7 @@ class TestPool(base.BaseAPITest):
         self.set_lb_status(self.lb_id)
         response = self.get(self.POOL_PATH.format(
             pool_id=api_pool.get('id'))).json.get(self.root_tag)
-        self.assertNotEqual('new_name', response.get('name'))
+        self.assertEqual('new_name', response.get('name'))
         self.assertIsNotNone(response.get('created_at'))
         self.assertIsNotNone(response.get('updated_at'))
         self.assert_correct_status(
@@ -839,6 +997,115 @@ class TestPool(base.BaseAPITest):
         self.assert_correct_lb_status(self.lb_id, constants.ONLINE,
                                       constants.ACTIVE)
 
+    def test_update_get_session_persistence_from_db_if_no_request(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0'}
+        optionals = {"listener_id": self.udp_listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            None,
+            constants.PROTOCOL_UDP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.udp_lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        ty = sess_p.pop('type')
+        sess_p['persistence_timeout'] = 4
+        sess_p['persistence_granularity'] = "255.255.0.0"
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool))
+        sess_p['type'] = ty
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        self.assertEqual(sess_p, response.get('session_persistence'))
+        self.assert_correct_status(
+            listener_id=self.udp_listener_id,
+            pool_id=api_pool.get('id'),
+            listener_prov_status=constants.PENDING_UPDATE,
+            pool_prov_status=constants.PENDING_UPDATE)
+
+    def test_update_udp_case_source_ip(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0'}
+        optionals = {"listener_id": self.udp_listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            None,
+            constants.PROTOCOL_UDP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.udp_lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        sess_p['persistence_timeout'] = 4
+        sess_p['persistence_granularity'] = "255.255.0.0"
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool))
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        self.assertEqual(sess_p, response.get('session_persistence'))
+        self.assert_correct_status(
+            listener_id=self.udp_listener_id,
+            pool_id=api_pool.get('id'),
+            listener_prov_status=constants.PENDING_UPDATE,
+            pool_prov_status=constants.PENDING_UPDATE)
+
+        self.set_lb_status(self.udp_lb_id)
+        self.set_object_status(self.pool_repo, api_pool.get('id'))
+        # Negative cases
+        # Error during update pool with non-UDP type and cookie_name.
+        expect_error_msg = (
+            "Validation failure: Cookie names are not supported for %s"
+            " pools.") % constants.PROTOCOL_UDP
+        sess_p['type'] = constants.SESSION_PERSISTENCE_HTTP_COOKIE
+        sess_p['cookie_name'] = 'test-cookie-name'
+        new_pool = {'session_persistence': sess_p}
+        res = self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                       self._build_body(new_pool), status=400,
+                       expect_errors=True)
+        self.assertEqual(expect_error_msg, res.json['faultstring'])
+        self.assert_correct_status(
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
+        # Error during update pool with source ip type and more options.
+        expect_error_msg = (
+            "Validation failure: session_persistence %s type for %s protocol "
+            "only accepts: type, persistence_timeout, "
+            "persistence_granularity.") % (
+            constants.SESSION_PERSISTENCE_SOURCE_IP, constants.PROTOCOL_UDP)
+        sess_p['type'] = constants.SESSION_PERSISTENCE_SOURCE_IP
+        sess_p['cookie_name'] = 'test-cookie-name'
+        sess_p['persistence_timeout'] = 4
+        sess_p['persistence_granularity'] = "255.255.0.0"
+        res = self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                       self._build_body(new_pool), status=400,
+                       expect_errors=True)
+        self.assertEqual(expect_error_msg, res.json['faultstring'])
+        self.assert_correct_status(
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
+        # Error during update pool with non-UDP session persistence type.
+        sess_p['cookie_name'] = None
+        for ty in [constants.SESSION_PERSISTENCE_APP_COOKIE,
+                   constants.SESSION_PERSISTENCE_HTTP_COOKIE]:
+            expect_error_msg = ("Validation failure: Session persistence of "
+                                "type %s is not supported for %s protocol "
+                                "pools.") % (ty, constants.PROTOCOL_UDP)
+            sess_p['type'] = ty
+            res = self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                           self._build_body(new_pool), status=400,
+                           expect_errors=True)
+            self.assertEqual(expect_error_msg, res.json['faultstring'])
+            self.assert_correct_status(
+                lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
+
     def test_bad_update(self):
         api_pool = self.create_pool(
             self.lb_id,
@@ -853,7 +1120,8 @@ class TestPool(base.BaseAPITest):
             lb_id=self.lb_id, listener_id=self.listener_id,
             pool_id=api_pool.get('id'))
 
-    def test_update_with_bad_handler(self):
+    @mock.patch('octavia.api.drivers.utils.call_provider')
+    def test_update_with_bad_provider(self, mock_provider):
         api_pool = self.create_pool(
             self.lb_id,
             constants.PROTOCOL_HTTP,
@@ -861,13 +1129,33 @@ class TestPool(base.BaseAPITest):
             listener_id=self.listener_id).get(self.root_tag)
         self.set_lb_status(lb_id=self.lb_id)
         new_pool = {'name': 'new_name'}
-        self.handler_mock().pool.update.side_effect = Exception()
-        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
-                 self._build_body(new_pool))
+        mock_provider.side_effect = exceptions.ProviderDriverError(
+            prov='bad_driver', user_msg='broken')
+        response = self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                            self._build_body(new_pool), status=500)
+        self.assertIn('Provider \'bad_driver\' reports error: broken',
+                      response.json.get('faultstring'))
+
+    def test_bad_update_non_udp_pool_with_udp_fields(self):
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            listener_id=self.listener_id).get(self.root_tag)
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
+              "persistence_timeout": 3,
+              "persistence_granularity": '255.255.255.0'}
+        self.set_lb_status(self.lb_id)
+        new_pool = {'session_persistence': sp}
+        expect_error_msg = ("Validation failure: persistence_timeout and "
+                            "persistence_granularity is only for %s "
+                            "protocol pools.") % constants.PROTOCOL_UDP
+        res = self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                       self._build_body(new_pool), status=400,
+                       expect_errors=True)
+        self.assertEqual(expect_error_msg, res.json['faultstring'])
         self.assert_correct_status(
-            lb_id=self.lb_id, listener_id=self.listener_id,
-            pool_id=api_pool.get('id'),
-            pool_prov_status=constants.ERROR)
+            lb_id=self.udp_lb_id, listener_id=self.udp_listener_id)
 
     def test_delete(self):
         api_pool = self.create_pool(
@@ -994,7 +1282,8 @@ class TestPool(base.BaseAPITest):
         self.delete(self.POOL_PATH.format(
             pool_id=api_pool.get('id')), status=409)
 
-    def test_delete_with_bad_handler(self):
+    @mock.patch('octavia.api.drivers.utils.call_provider')
+    def test_delete_with_bad_provider(self, mock_provider):
         api_pool = self.create_pool(
             self.lb_id,
             constants.PROTOCOL_HTTP,
@@ -1010,15 +1299,13 @@ class TestPool(base.BaseAPITest):
         self.assertIsNone(api_pool.pop('updated_at'))
         self.assertIsNotNone(response.pop('updated_at'))
         self.assertEqual(api_pool, response)
-        self.handler_mock().pool.delete.side_effect = Exception()
-        self.delete(self.POOL_PATH.format(pool_id=api_pool.get('id')))
-        self.assert_correct_status(
-            lb_id=self.lb_id, listener_id=self.listener_id,
-            pool_id=api_pool.get('id'),
-            pool_prov_status=constants.ERROR)
+        mock_provider.side_effect = exceptions.ProviderDriverError(
+            prov='bad_driver', user_msg='broken')
+        self.delete(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                    status=500)
 
     def test_create_with_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
               "cookie_name": "test_cookie_name"}
         optionals = {"listener_id": self.listener_id,
                      "session_persistence": sp}
@@ -1039,7 +1326,7 @@ class TestPool(base.BaseAPITest):
             pool_id=api_pool.get('id'))).json.get(self.root_tag)
         sess_p = response.get('session_persistence')
         self.assertIsNotNone(sess_p)
-        self.assertEqual(constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        self.assertEqual(constants.SESSION_PERSISTENCE_APP_COOKIE,
                          sess_p.get('type'))
         self.assertEqual('test_cookie_name', sess_p.get('cookie_name'))
         self.assert_correct_status(
@@ -1057,9 +1344,54 @@ class TestPool(base.BaseAPITest):
             'session_persistence': sp}
         self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
 
-    def test_add_session_persistence(self):
+    def test_create_with_bad_SP_type_HTTP_cookie(self):
         sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
               "cookie_name": "test_cookie_name"}
+        lb_pool = {
+            'loadbalancer_id': self.lb_id,
+            'listener_id': self.listener_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
+
+    def test_create_with_bad_SP_type_IP_cookie(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP,
+              "cookie_name": "test_cookie_name"}
+        lb_pool = {
+            'loadbalancer_id': self.lb_id,
+            'listener_id': self.listener_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
+
+    def test_create_with_bad_SP_cookie_name(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
+              "cookie_name": "b@d_cookie_name"}
+        lb_pool = {
+            'loadbalancer_id': self.lb_id,
+            'listener_id': self.listener_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
+
+    def test_create_with_missing_cookie_name(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE}
+        lb_pool = {
+            'loadbalancer_id': self.lb_id,
+            'listener_id': self.listener_id,
+            'protocol': constants.PROTOCOL_HTTP,
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'session_persistence': sp}
+        self.post(self.POOLS_PATH, self._build_body(lb_pool), status=400)
+
+    def test_add_session_persistence(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
+              "cookie_name": "test_cookie_name",
+              'persistence_granularity': None,
+              'persistence_timeout': None}
         api_pool = self.create_pool(
             self.lb_id,
             constants.PROTOCOL_HTTP,
@@ -1071,7 +1403,7 @@ class TestPool(base.BaseAPITest):
                  self._build_body(new_pool))
         response = self.get(self.POOL_PATH.format(
             pool_id=api_pool.get('id'))).json.get(self.root_tag)
-        self.assertNotEqual(sp, response.get('session_persistence'))
+        self.assertEqual(sp, response.get('session_persistence'))
         self.assert_correct_status(
             lb_id=self.lb_id, listener_id=self.listener_id,
             pool_id=api_pool.get('id'),
@@ -1080,7 +1412,7 @@ class TestPool(base.BaseAPITest):
             pool_prov_status=constants.PENDING_UPDATE)
 
     def test_update_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
               "cookie_name": "test_cookie_name"}
         optionals = {"listener_id": self.listener_id,
                      "session_persistence": sp}
@@ -1100,7 +1432,7 @@ class TestPool(base.BaseAPITest):
                  self._build_body(new_pool))
         response = self.get(self.POOL_PATH.format(
             pool_id=api_pool.get('id'))).json.get(self.root_tag)
-        self.assertNotEqual(sess_p, response.get('session_persistence'))
+        self.assertEqual(sess_p, response.get('session_persistence'))
         self.assert_correct_status(
             lb_id=self.lb_id, listener_id=self.listener_id,
             pool_id=api_pool.get('id'),
@@ -1109,8 +1441,10 @@ class TestPool(base.BaseAPITest):
             pool_prov_status=constants.PENDING_UPDATE)
 
     def test_update_preserve_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
-              "cookie_name": "test_cookie_name"}
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
+              "cookie_name": "test_cookie_name",
+              'persistence_granularity': None,
+              'persistence_timeout': None}
         optionals = {"listener_id": self.listener_id,
                      "name": "name", "session_persistence": sp}
         api_pool = self.create_pool(
@@ -1133,7 +1467,7 @@ class TestPool(base.BaseAPITest):
             pool_prov_status=constants.PENDING_UPDATE)
 
     def test_update_bad_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
               "cookie_name": "test_cookie_name"}
         optionals = {"listener_id": self.listener_id,
                      "session_persistence": sp}
@@ -1151,8 +1485,83 @@ class TestPool(base.BaseAPITest):
         self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
                  self._build_body(new_pool), status=400)
 
+    def test_update_with_bad_SP_type_HTTP_cookie(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP}
+        optionals = {"listener_id": self.listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        sess_p['type'] = constants.SESSION_PERSISTENCE_HTTP_COOKIE
+        sess_p['cookie_name'] = 'test_cookie_name'
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool), status=400)
+
+    def test_update_with_bad_SP_type_IP_cookie(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE}
+        optionals = {"listener_id": self.listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        sess_p['type'] = constants.SESSION_PERSISTENCE_SOURCE_IP
+        sess_p['cookie_name'] = 'test_cookie_name'
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool), status=400)
+
+    def test_update_with_bad_SP_cookie_name(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP}
+        optionals = {"listener_id": self.listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        sess_p['type'] = constants.SESSION_PERSISTENCE_APP_COOKIE
+        sess_p['cookie_name'] = 'b@d_cookie_name'
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool), status=400)
+
+    def test_update_with_missing_SP_cookie_name(self):
+        sp = {"type": constants.SESSION_PERSISTENCE_SOURCE_IP}
+        optionals = {"listener_id": self.listener_id,
+                     "session_persistence": sp}
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            **optionals).get(self.root_tag)
+        self.set_lb_status(lb_id=self.lb_id)
+        response = self.get(self.POOL_PATH.format(
+            pool_id=api_pool.get('id'))).json.get(self.root_tag)
+        sess_p = response.get('session_persistence')
+        sess_p['type'] = constants.SESSION_PERSISTENCE_APP_COOKIE
+        new_pool = {'session_persistence': sess_p}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool), status=400)
+
     def test_delete_with_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
               "cookie_name": "test_cookie_name"}
         optionals = {"listener_id": self.listener_id,
                      "session_persistence": sp}
@@ -1171,7 +1580,7 @@ class TestPool(base.BaseAPITest):
             pool_prov_status=constants.PENDING_DELETE)
 
     def test_delete_session_persistence(self):
-        sp = {"type": constants.SESSION_PERSISTENCE_HTTP_COOKIE,
+        sp = {"type": constants.SESSION_PERSISTENCE_APP_COOKIE,
               "cookie_name": "test_cookie_name"}
         optionals = {"listener_id": self.listener_id,
                      "session_persistence": sp}
@@ -1184,7 +1593,7 @@ class TestPool(base.BaseAPITest):
         new_sp = {"pool": {"session_persistence": None}}
         response = self.put(self.POOL_PATH.format(
             pool_id=api_pool.get('id')), new_sp).json.get(self.root_tag)
-        self.assertIsNotNone(response.get('session_persistence'))
+        self.assertIsNone(response.get('session_persistence'))
         self.assert_correct_status(
             lb_id=self.lb_id, listener_id=self.listener_id,
             pool_id=api_pool.get('id'),
@@ -1264,6 +1673,18 @@ class TestPool(base.BaseAPITest):
         self.delete(self.POOL_PATH.format(pool_id=api_pool.get('id')),
                     status=409)
 
+    def test_update_already_deleted(self):
+        api_pool = self.create_pool(
+            self.lb_id,
+            constants.PROTOCOL_HTTP,
+            constants.LB_ALGORITHM_ROUND_ROBIN,
+            listener_id=self.listener_id).get(self.root_tag)
+        # This updates the child objects
+        self.set_lb_status(self.lb_id, status=constants.DELETED)
+        new_pool = {'admin_state_up': False}
+        self.put(self.POOL_PATH.format(pool_id=api_pool.get('id')),
+                 self._build_body(new_pool), status=404)
+
     def test_delete_already_deleted(self):
         api_pool = self.create_pool(
             self.lb_id,
@@ -1273,4 +1694,59 @@ class TestPool(base.BaseAPITest):
         # This updates the child objects
         self.set_lb_status(self.lb_id, status=constants.DELETED)
         self.delete(self.POOL_PATH.format(pool_id=api_pool.get('id')),
-                    status=204)
+                    status=404)
+
+    @mock.patch('octavia.common.tls_utils.cert_parser.load_certificates_data')
+    def test_valid_listener_pool_protocol(self, mock_cert_data):
+        cert = data_models.TLSContainer(certificate='cert')
+        lb_pool = {
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'project_id': self.project_id}
+        mock_cert_data.return_value = {'sni_certs': [cert]}
+        valid_map = constants.VALID_LISTENER_POOL_PROTOCOL_MAP
+        port = 1
+        for listener_proto in valid_map:
+            for pool_proto in valid_map[listener_proto]:
+                port = port + 1
+                opts = {}
+                if listener_proto == constants.PROTOCOL_TERMINATED_HTTPS:
+                    opts['sni_container_refs'] = [uuidutils.generate_uuid()]
+                listener = self.create_listener(
+                    listener_proto, port, self.lb_id, **opts).get('listener')
+                self.set_object_status(self.lb_repo, self.lb_id)
+                if listener['default_pool_id'] is None:
+                    lb_pool['protocol'] = pool_proto
+                    lb_pool['listener_id'] = listener.get('id')
+                    self.post(self.POOLS_PATH, self._build_body(lb_pool),
+                              status=201)
+                    self.set_object_status(self.lb_repo, self.lb_id)
+
+    @mock.patch('octavia.common.tls_utils.cert_parser.load_certificates_data')
+    def test_invalid_listener_pool_protocol_map(self, mock_cert_data):
+        cert = data_models.TLSContainer(certificate='cert')
+        lb_pool = {
+            'lb_algorithm': constants.LB_ALGORITHM_ROUND_ROBIN,
+            'project_id': self.project_id}
+        mock_cert_data.return_value = {'sni_certs': [cert]}
+        invalid_map = c_const.INVALID_LISTENER_POOL_PROTOCOL_MAP
+        port = 1
+        for listener_proto in invalid_map:
+            opts = {}
+            if listener_proto == constants.PROTOCOL_TERMINATED_HTTPS:
+                opts['sni_container_refs'] = [uuidutils.generate_uuid()]
+            listener = self.create_listener(
+                listener_proto, port, self.lb_id, **opts).get('listener')
+            self.set_object_status(self.lb_repo, self.lb_id)
+            port = port + 1
+            for pool_proto in invalid_map[listener_proto]:
+                expect_error_msg = ("Validation failure: The pool protocol "
+                                    "'%s' is invalid while the listener "
+                                    "protocol is '%s'.") % (pool_proto,
+                                                            listener_proto)
+                if listener['default_pool_id'] is None:
+                    lb_pool['protocol'] = pool_proto
+                    lb_pool['listener_id'] = listener.get('id')
+                    res = self.post(self.POOLS_PATH, self._build_body(lb_pool),
+                                    status=400, expect_errors=True)
+                    self.assertEqual(expect_error_msg, res.json['faultstring'])
+                    self.assert_correct_status(lb_id=self.lb_id)
